import argparse

from fan_tools.unix import wait_no_socket, wait_socket

parser = argparse.ArgumentParser(description='Wait for socket')

parser.add_argument('-t', '--timeout', type=int, default=120)
parser.add_argument('-n', '--no-socket', action='store_true', default=False)
parser.add_argument('host', type=str)
parser.add_argument('port', type=int)


def main():
    args = parser.parse_args()
    print(args)
    f = wait_no_socket if args.no_socket else wait_socket
    if not f(args.host, args.port, args.timeout):
        print('Timeout')
        exit(1)
    else:
        print('Ok')


if __name__ == '__main__':
    main()



from django.db.models import Index


class LTreeIndex(Index):

    def create_sql(self, model, schema_editor, using=''):
        """
        Update ltree label path when parent/child relationships changed
        Taken from: https://github.com/peopledoc/django-ltree-demo
        """

        return """
        -- function to calculate the path of any given node
        CREATE OR REPLACE FUNCTION _update_{db_table}_path() RETURNS TRIGGER AS
        $$
        BEGIN
            IF NEW.parent_id IS NULL THEN
                NEW.ltree_label_path = NEW.ltree_label::ltree;
            ELSE
                SELECT ltree_label_path || NEW.ltree_label
                    FROM {db_table}
                    WHERE NEW.parent_id IS NULL or id = NEW.parent_id
                    INTO NEW.ltree_label_path;
            END IF;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;


        -- function to update the path of the descendants of a node
        CREATE OR REPLACE FUNCTION _update_descendants_{db_table}_path() RETURNS TRIGGER AS
        $$
        BEGIN
            UPDATE {db_table}
                SET ltree_label_path =
                    NEW.ltree_label_path ||
                    subpath({db_table}.ltree_label_path, nlevel(OLD.ltree_label_path))
                WHERE {db_table}.ltree_label_path <@ OLD.ltree_label_path AND id != NEW.id;
            RETURN NEW;
        END;
        $$ LANGUAGE plpgsql;


        -- calculate the path every time we insert a new node
        DROP TRIGGER IF EXISTS {db_table}_path_insert_trg ON {db_table};
        CREATE TRIGGER {db_table}_path_insert_trg
            BEFORE INSERT ON {db_table}
            FOR EACH ROW
            EXECUTE PROCEDURE _update_{db_table}_path();


        -- calculate the path when updating the parent or the label
        DROP TRIGGER IF EXISTS {db_table}_path_update_trg ON {db_table};
        CREATE TRIGGER {db_table}_path_update_trg
            BEFORE UPDATE ON {db_table}
            FOR EACH ROW
            WHEN (OLD.parent_id IS DISTINCT FROM NEW.parent_id
                OR OLD.ltree_label IS DISTINCT FROM NEW.ltree_label)
            EXECUTE PROCEDURE _update_{db_table}_path();


        -- if the path was updated, update the path of the descendants
        DROP TRIGGER IF EXISTS {db_table}_path_after_trg ON {db_table};
        CREATE TRIGGER {db_table}_path_after_trg
            AFTER UPDATE ON {db_table}
            FOR EACH ROW
            WHEN (NEW.ltree_label_path IS DISTINCT FROM OLD.ltree_label_path)
            EXECUTE PROCEDURE _update_descendants_{db_table}_path();
        """.format(
            db_table=model._meta.db_table,
        )



def assert_validation_error(response, field, code):
    assert 'error' in response
    error = response['error']

    assert 'detail' in error
    detail = error['detail']

    assert field in detail
    items = detail[field]

    found = False
    for item in items:
        if item['code'] == code:
            found = True

    assert found



import os
import sys
from decimal import ROUND_HALF_UP, Decimal

import py


def execfile(fname, _globals, _locals):
    """
    Usage: execfile('path/to/file.py', globals(), locals())
    """
    if os.path.exists(fname):
        with open(fname) as f:
            code = compile(f.read(), os.path.basename(fname), 'exec')
            exec(code, _globals, _locals)
            return True
    else:
        return False


def rel_path(path, check=False, depth=1):
    d = os.path.dirname(sys._getframe(depth).f_code.co_filename)
    full = os.path.abspath(os.path.join(d, path))
    if check and not os.path.exists(full):
        raise Exception('No such path: {!r}'.format(full))
    return full


def py_rel_path(*args, **kwargs):
    return py.path.local(rel_path(*args, depth=2, **kwargs))


def auto_directory(rel_name):
    """
    if you're using py.path you make do that as:
    py.path.local(full_path).ensure_dir()
    """
    dir_name = rel_path(rel_name, check=False)
    if not os.path.exists(dir_name):
        os.makedirs(dir_name, exist_ok=True)
    return dir_name


def usd_round(amount):
    if isinstance(amount, str):
        amount = Decimal(amount)
    return amount.quantize(Decimal('.01'), rounding=ROUND_HALF_UP)



from django.db import models
from django.utils.text import slugify
from django.contrib.postgres.indexes import GistIndex, Index

from .fields import LTreeLabelField, LTreeLabelPathField
from .indexes import LTreeIndex


class LTreeModel(models.Model):

    ltree_defaut_label_field = 'name'

    ltree_label = LTreeLabelField()
    ltree_label_path = LTreeLabelPathField()

    parent = models.ForeignKey(
        'self',
        null=True,
        blank=True,
        related_name='children',
        on_delete=models.CASCADE,
    )

    class Meta:
        abstract = True

        indexes = [
            Index(fields=['ltree_label_path']),
            GistIndex(fields=['ltree_label_path']),
            LTreeIndex(fields=['ltree_label_path']),
        ]

    @property
    def depth(self):
        return len(
            self.ltree_label_path.split('.'),
        )

    def save(self, *args, **kwargs):
        if not self.ltree_label:
            self.ltree_label = slugify(
                getattr(
                    self,
                    self.ltree_defaut_label_field
                ),
            ).replace('-', '_')
        super().save(*args, **kwargs)



import asyncio
import logging

import gitlab

LOOP = 20
log = logging.getLogger('mon_server.gitlab_runners')


def update_gitlab_runners(gitlab_api):
    metrics = {}
    try:
        log.debug('Going to update runners state')
        runners = gitlab_api.runners.all(scope='online')
        for runner in runners:
            if runner['description'].endswith('disabled'):
                log.debug(f'Skip {runner}')
                continue
            elif runner['active']:
                k = 'ci_runner{config="%s"}' % runner['description']
                metrics[k] = 1
        log.debug(f'State was updated to {metrics}')
    except Exception:
        log.exception('During update_gitlab_runners')
    return metrics


async def update_gitlab_loop(update_metrics, params):
    """
    app = Sanic()
    mserver = MetricsServer(app)
    mserver.add_task(update_gitlab_loop, params={'url': GITLAB_URL, 'token': token})
    """
    gitlab_api = gitlab.Gitlab(url=params['url'], private_token=params['token'], api_version=4)
    while True:
        try:
            metrics = update_gitlab_runners(gitlab_api)
            update_metrics(metrics)
        except Exception:
            update_metrics({})
            log.exception('During loop')
        await asyncio.sleep(LOOP)



from django.db import transaction

_global = {'request_id': None, 'data': {}}


def _get_request_unique_cache():
    try:
        import uwsgi

        rid = uwsgi.request_id()
        if rid != _global['request_id']:
            _global['request_id'] = rid
            _global['data'] = {}
        data = _global['data']
    except:
        data = {}

    return data


def request_uniq(func):
    """
    return unique dict for each uwsgi request.
    note: won't work on non-uwsgi cases
    """

    def _wrapped(*args, **kwargs):
        data = _get_request_unique_cache()
        return func(data, *args, **kwargs)

    return _wrapped


def _deffered_on_commit():
    return transaction.on_commit


def on_commit():
    return _deffered_on_commit()


def call_once_on_commit(func):
    key = '{}.{}_is_added_on_commit'.format(func.__module__, func.__name__)

    @request_uniq
    def _wrapped(cache, *args, **kwargs):
        if cache.get(key) is None:
            on_commit()(lambda: func(*args, **kwargs))
            cache[key] = True

    return _wrapped



import logging
from functools import reduce

import django_filters
from django.db.models import Q, F, Value
from django_filters import CharFilter
from django_filters.rest_framework import BaseInFilter, NumberFilter

from fan_tools.django.db.pgfields import WordSimilarity
from fan_tools.django.db.utils import set_word_similarity_threshold


class NumberInFilter(BaseInFilter, NumberFilter):
    pass


class TrigramFilter(CharFilter):
    def __init__(self, *args, **kwargs):
        self.similarity_threshold = kwargs.pop('similarity_threshold', None)
        super().__init__(*args, **kwargs)

    def filter(self, qs, value):
        if value:
            if self.similarity_threshold:
                set_word_similarity_threshold(1 - self.similarity_threshold)
            similarity = WordSimilarity(Value(value), F(self.field_name))
            qs = (
                qs.annotate(similarity=similarity)
                .order_by('similarity', 'pk')
                .filter(**{f'{self.field_name}__similar': value})
            )
        return qs


class EnumFilter(django_filters.CharFilter):
    def __init__(self, enum, *args, **kwargs):
        self._enum = enum
        super().__init__(*args, **kwargs)

    def filter(self, qs, name):
        # In django filters 2.0 `name` is renamed to `field_name`
        self_name = getattr(self, 'name', None) or getattr(self, 'field_name')
        try:
            q_objects = []
            if isinstance(name, str):
                names = name.split(',')
                for n in names:
                    if hasattr(self._enum, n):
                        value = getattr(self._enum, n)
                        q = Q(**{'{}__exact'.format(self_name): value})
                        q_objects.append(q)
                    elif n == 'null':
                        q_objects.append(Q(**{'{}__isnull'.format(self_name): True}))
            elif name is None:
                q_objects.append(Q(**{'{}__isnull'.format(self_name): True}))
            else:
                raise AttributeError
            if q_objects:
                return self.get_method(qs)(reduce(lambda q1, q2: q1 | q2, q_objects))
            else:
                return qs

        except Exception:
            logging.exception('Failed to convert value: {} {}'.format(self_name, name))
            return super(EnumFilter, self).filter(qs, None)



import inspect
import uuid
from typing import Callable, Dict, Optional, Type, TypeVar, Union, Any

from django.http import QueryDict, HttpRequest
from rest_framework import serializers, exceptions
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.views import APIView


_ValidationSchema = Union[
    Dict[str, serializers.Field],
    Type[serializers.Serializer]
]

_C = TypeVar('_C', bound=Callable)


def make_serializer_cls(fields_mapping: Dict[str, serializers.Field]) -> Type[serializers.Serializer]:
    cls_name = uuid.uuid4().hex + 'AnonymousSerializer'
    bases = (serializers.Serializer,)
    # noinspection PyTypeChecker
    return type(cls_name, bases, fields_mapping)


def validate_with_serializer(serializer_cls: Type[serializers.Serializer], data: Any) -> Any:
    serializer = serializer_cls(data=data)
    try:
        serializer.is_valid(raise_exception=True)
        return serializer.validated_data

    except serializers.ValidationError as exc:
        raise exceptions.ParseError(exc.detail)


def validate_request_data(serializer_cls: Type[serializers.Serializer],
                          request: Union[Request, HttpRequest],
                          **top_kwargs) -> Any:
    def validated_form(request, **kwargs):
        data = request.query_params.dict() if request.method in ['GET'] else request.data
        if isinstance(data, QueryDict):
            form = serializer_cls(data={**data.dict(), **kwargs})
        elif isinstance(data, dict):
            form = serializer_cls(data={**data, **kwargs})
        else:
            form = serializer_cls(data=data, **kwargs)
        form.is_valid(raise_exception=True)
        return form

    if request:
        kwargs = {}
        if request.resolver_match:
            kwargs = {**request.resolver_match.kwargs}
        if top_kwargs:
            kwargs = {**kwargs, **top_kwargs}

        return validated_form(request, **kwargs).validated_data


class validate_request:
    def __init__(self, *,
                 query_params: Optional[_ValidationSchema] = None,
                 data: Optional[_ValidationSchema] = None
                 ) -> None:
        if isinstance(query_params, dict):
            query_params = make_serializer_cls(query_params)
        self.query_params_serializer_cls = query_params

        if isinstance(data, dict):
            data = make_serializer_cls(data)
        self.data_serializer_cls = data

    def __call__(self, method: _C) -> _C:
        _self = self

        if not inspect.getfullargspec(method).args or not inspect.getfullargspec(method).args[0] == 'self':
            # relying on convention to determine whether it's a method or a function
            raise TypeError('@validate_request cannot be used on function-based views')

        def method_wrapper(self: APIView, request: Request, *args: Any, **kwargs: Any) -> Response:
            if not isinstance(self, APIView):
                raise TypeError('@validate_request can only be used on methods of APIView')

            if hasattr(self, 'log'):
                if _self.query_params_serializer_cls:
                    _self.query_params_serializer_cls.log = self.log
                if _self.data_serializer_cls:
                    _self.data_serializer_cls.log = self.log

            if _self.query_params_serializer_cls is not None:
                self.validated_query_params = validate_with_serializer(_self.query_params_serializer_cls, request.query_params)

            if _self.data_serializer_cls is not None:
                self.validated_data = validate_with_serializer(_self.data_serializer_cls, request.data)

            return method(self, request, *args, **kwargs)

        return method_wrapper
